/*
 * Copyright (C) 2000-2001 Damien Douxchamps  <douxchamps@ieee.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include "callback_proc.h"
#include "support.h"
#include "update_ranges.h"
#include "definitions.h"
#include "conversions.h"
#include <gdk/gdkprivate.h>
#include <libdc1394/dc1394_control.h>

extern GtkWidget *porthole_window;
extern char* feature_op_list[NUM_FEATURES];
extern char* feature_scale_list[NUM_FEATURES];
extern dc1394_camerainfo *camera;
extern dc1394_feature_set *feature_set;
extern dc1394_miscinfo *misc_info;
extern porthole_info pi;
extern dc1394_cameracapture *capture;
extern guint gIdleID;

/* globals required to communicate with iso handler */
unsigned char *g_buffer_pos;
size_t         g_packet_length;
unsigned char	g_new_frame;

void OpRangeProcedure(GtkButton* widget, int feature)
{
  int err, timeout;
  dc1394bool_t value=TRUE;
  timeout=LOOP_RETRIES;
  err=dc1394_start_one_push_operation(camera->handle, camera->id, feature);
  gtk_widget_set_sensitive(GTK_WIDGET (widget), FALSE);
  while ( value & (timeout>0) )
    {
      usleep(LOOP_SLEEP);
      err=dc1394_is_one_push_in_operation(camera->handle, camera->id, feature, &value);
      // the next line is not working: only the last update occurs
      UpdateRangeValue(GTK_WIDGET (widget),feature);
      timeout--;
    }
  if (timeout==0)
    printf("OP function timed-out!\n");
  gtk_widget_set_sensitive(GTK_WIDGET (widget),TRUE);

}

void ManRangeProcedure(GtkToggleButton *widget, int feature)
{
  int err;

  if (widget->active)
    UpdateRangeValue(GTK_WIDGET (widget),feature);

  // sensitive OP function if available and in man mode:
  gtk_widget_set_sensitive(GTK_WIDGET (lookup_widget(GTK_WIDGET (widget), feature_op_list[feature])),
			   feature_set->feature[feature].one_push && widget->active);

  // sensitive ranges accoring to auto/man mode and update value:
  switch (feature)
    {
      case FEATURE_WHITE_BALANCE:
	gtk_widget_set_sensitive(GTK_WIDGET (lookup_widget(GTK_WIDGET (widget), "whitebal_RV_scale")),widget->active);
	gtk_widget_set_sensitive(GTK_WIDGET (lookup_widget(GTK_WIDGET (widget), "whitebal_BU_scale")),widget->active);break;
      case FEATURE_TEMPERATURE:
	gtk_widget_set_sensitive(GTK_WIDGET (lookup_widget(GTK_WIDGET (widget), "temp_goal_scale")),widget->active);break;
	// only the goal range is active, the other one is just an indicator
      default:
	gtk_widget_set_sensitive(GTK_WIDGET (lookup_widget(GTK_WIDGET (widget), feature_scale_list[feature])),widget->active);
    }

  err=dc1394_auto_on_off(camera->handle, camera->id, feature, !widget->active);
  feature_set->feature[feature].auto_active=!widget->active;

}


void PowerRangeProcedure(GtkToggleButton* widget, int feature)
{
  int err;
  err=dc1394_feature_on_off(camera->handle, camera->id, feature, widget->active);
  feature_set->feature[feature].is_on=widget->active;

  switch(feature)
    {
      case FEATURE_WHITE_BALANCE:
	gtk_widget_set_sensitive(GTK_WIDGET (lookup_widget(GTK_WIDGET (widget), "whitebal_RV_scale")),widget->active);
	gtk_widget_set_sensitive(GTK_WIDGET (lookup_widget(GTK_WIDGET (widget), "whitebal_BU_scale")),widget->active);break;
      case FEATURE_TEMPERATURE:
	gtk_widget_set_sensitive(GTK_WIDGET (lookup_widget(GTK_WIDGET (widget), "temp_goal_scale")),widget->active);break;
      default:
	gtk_widget_set_sensitive(GTK_WIDGET (lookup_widget(GTK_WIDGET (widget), feature_scale_list[feature])),widget->active);
    }

}

void IsoFlowCheck()
{ 
  int err;
  err=dc1394_get_iso_channel_and_speed(camera->handle, camera->id, &misc_info->iso_channel, &misc_info->iso_speed);
  err=dc1394_get_iso_status(camera->handle, camera->id, &misc_info->is_iso_on);
  if (misc_info->is_iso_on)
    err=dc1394_stop_iso_transmission(camera->handle, camera->id);// if not done, restarting is no more possible
}

void IsoFlowResume()
{
  int err;
  GtkWidget *window = lookup_widget(porthole_window, "camera_scope");
  gboolean isPortholeOpen = (pi.handle != NULL);

  if (isPortholeOpen) 
  {
     gtk_idle_remove(gIdleID);
     IsoStopThread();
  }

  if (misc_info->is_iso_on)// restart if it was 'on' before the changes
  {
      err=dc1394_set_iso_channel_and_speed(camera->handle, camera->id, misc_info->iso_channel, misc_info->iso_speed);
      err=dc1394_start_iso_transmission(camera->handle, camera->id);
  }

  if (isPortholeOpen && IsoStartThread(window)>0 )
    // run a gtk idler thread to update the display
    gIdleID = gtk_idle_add_priority(GTK_PRIORITY_LOW, idler, NULL);
  
}

void* IsoReceiveThread(void* arg)
{
  porthole_info *p = (porthole_info *)arg;
  
  while (1) {
    pthread_testcancel();

    if (p->receive_method == RECEIVE_METHOD_RAW1394)
      dc1394_single_capture( p->handle, capture);
    else dc1394_dma_single_capture( capture);

    pthread_testcancel();
    pthread_mutex_lock(&(p->mutex));

#ifdef HAVE_X11_EXTENSIONS_XVLIB_H
    if (p->display_method == DISPLAY_METHOD_XV)
    {
      switch(misc_info->mode)
      {
         case MODE_160x120_YUV444:
           iyu22yuy2( (char *) capture->capture_buffer, p->xv_image->data,
             capture->frame_width*capture->frame_height);
           break;
         case MODE_640x480_YUV422:
         case MODE_320x240_YUV422:
           if (p->xv_format == GUID_UYVY_PACKED)
             memcpy( p->xv_image->data, (char *) capture->capture_buffer, 
             capture->quadlets_per_frame*4);
           else
             uyvy2yuy2( (char *) capture->capture_buffer, p->xv_image->data, 
             capture->frame_width*capture->frame_height);
           break;
         case MODE_640x480_YUV411:
           iyu12yuy2( (char *) capture->capture_buffer, p->xv_image->data, 
             capture->frame_width*capture->frame_height);
           break;
         case MODE_640x480_RGB:
           rgb2yuy2( (char *) capture->capture_buffer, p->xv_image->data, 
             capture->frame_width*capture->frame_height);
           break;
         case MODE_640x480_MONO:
           y2yuy2( (char *) capture->capture_buffer, p->xv_image->data, 
             capture->frame_width*capture->frame_height);
           break;
         default:
           return NULL;
      };
    } 
    else
#endif
    { //DISPLAY_METHOD_GDK
      switch(misc_info->mode)
      {
         case MODE_160x120_YUV444:
           iyu22rgb( (char *) capture->capture_buffer, p->gdk_buffer,
             capture->frame_width*capture->frame_height);
           break;
         case MODE_640x480_YUV422:
         case MODE_320x240_YUV422:
           uyvy2rgb( (char *) capture->capture_buffer, p->gdk_buffer,
             capture->frame_width*capture->frame_height);
           break;
         case MODE_640x480_YUV411:
           iyu12rgb( (char *) capture->capture_buffer, p->gdk_buffer,
             capture->frame_width*capture->frame_height);
           break;
         case MODE_640x480_RGB:
           memcpy( p->gdk_buffer, capture->capture_buffer,
             capture->quadlets_per_frame*4);
           break;
         case MODE_640x480_MONO:
           y2rgb( (char *) capture->capture_buffer, p->gdk_buffer,
             capture->frame_width*capture->frame_height);
           break;
         default:
           return NULL;
      };

    }
    p->buffer_is_dirty = 1;
    pthread_mutex_unlock(&(p->mutex));
    pthread_testcancel();
    if (p->receive_method == RECEIVE_METHOD_VIDEO1394)
        dc1394_dma_done_with_buffer(capture);
  }
  return NULL;
}

gint IsoStartThread(gpointer p)
{
  GtkWidget *scope = lookup_widget( GTK_WIDGET(p), "camera_scope");
  /* currently only supports VGA NONCOMPRESSED */
  if (misc_info->format > 0) return(-1);

  pi.handle = NULL;
  pi.gdk_buffer = NULL;
  pi.drawable = (GtkWidget *)p;
  pi.buffer_is_dirty = 0;
  pthread_mutex_init(&pi.mutex, NULL);

  // the iso receive handler gets its own raw1394 handle to free the controls
  if ( (pi.handle = dc1394_create_handle(0)) < 0)
  {
    return(-1);
  }
  
  if (dc1394_dma_setup_capture( camera->handle, camera->id, misc_info->iso_channel, 
    misc_info->format, misc_info->mode, misc_info->iso_speed, misc_info->framerate,
    4, capture) == DC1394_SUCCESS)
      pi.receive_method = RECEIVE_METHOD_VIDEO1394;
  else 
  if (dc1394_setup_capture( camera->handle, camera->id, misc_info->iso_channel, 
    misc_info->format, misc_info->mode, misc_info->iso_speed, misc_info->framerate, capture)
    == DC1394_SUCCESS)
      pi.receive_method = RECEIVE_METHOD_RAW1394;
  else {
    raw1394_destroy_handle( pi.handle);
    return(-1);
  }
#ifdef HAVE_X11_EXTENSIONS_XVLIB_H
  if (xvInit())
  {
    pi.display_method = DISPLAY_METHOD_XV;
  }
  else
#endif
  { 
    pi.display_method = DISPLAY_METHOD_GDK;
    pi.gdk_buffer = malloc( capture->frame_width * capture->frame_height * 3);
  }
  gtk_widget_set_usize( scope, capture->frame_width, capture->frame_height);

  pthread_create(&pi.thread, NULL, IsoReceiveThread, &pi);

  return (1);
}

gint IsoStopThread(void)
{
  pthread_mutex_lock(&pi.mutex);
#ifdef HAVE_X11_EXTENSIONS_XVLIB_H
  if (pi.xv_image != NULL)
  {
    XvStopVideo( pi.display, pi.xv_port, pi.window);
    XShmDetach( pi.display, &pi.xv_shm_info);
    shmdt( pi.xv_shm_info.shmaddr);
    shmctl( pi.xv_shm_info.shmid, IPC_RMID, 0);
    XFree( pi.xv_image);
  }
#endif
  
  if (pi.gdk_buffer != NULL) free(pi.gdk_buffer);
  
  pthread_cancel(pi.thread);
  pthread_mutex_unlock(&pi.mutex);
  pthread_join(pi.thread, NULL);
  if (pi.handle != NULL)
  {
    if (pi.receive_method == RECEIVE_METHOD_VIDEO1394)
      dc1394_dma_release_camera(camera->handle, capture);
    else dc1394_release_camera(camera->handle, capture);
    raw1394_destroy_handle(pi.handle);
    pi.handle = NULL;
  }
  return (1);
}

#ifdef HAVE_X11_EXTENSIONS_XVLIB_H

gint xvInit(void)
{
  int num_adaptors;
  int num_formats;
  XvImageFormatValues *formats;
  int i,j;
  char xv_name[5];
  XvAdaptorInfo	*adaptor_info;
  XGCValues xgcv;

  GdkWindowPrivate *priv = (GdkWindowPrivate*) pi.drawable->window;
  pi.display = priv->xdisplay;
  pi.window = priv->xwindow;

  pi.xv_port = -1;
  XvQueryAdaptors( pi.display, DefaultRootWindow(pi.display), &num_adaptors, &adaptor_info);

  for(i=0;i<num_adaptors && pi.xv_port < 0;i++)
  {
    formats = XvListImageFormats( pi.display, adaptor_info[i].base_id, &num_formats);
    for (j=0; j<num_formats && pi.xv_port < 0; j++)
    {
      xv_name[4]=0;
      memcpy( xv_name, &formats[j].id, 4);
      if ( (misc_info->mode == MODE_640x480_YUV422 || misc_info->mode == MODE_320x240_YUV422) &&
        formats[j].id ==  GUID_UYVY_PACKED) 
      {
        pi.xv_port = adaptor_info[i].base_id;
        pi.xv_format = GUID_UYVY_PACKED;
      } else if (formats[j].id ==  GUID_YUY2_PACKED)
      {
        pi.xv_port = adaptor_info[i].base_id;
        pi.xv_format = GUID_YUY2_PACKED;
      }
    }
    XFree(formats);
  }
  
  if (pi.xv_port < 0)
  {
    fprintf(stderr,"No suitable Xv adaptor found, resorting to GDK\n");	
    return(0);
  } else {
    pi.xv_image = (XvImage *) XvShmCreateImage( pi.display, pi.xv_port,
    pi.xv_format, 0, capture->frame_width, capture->frame_height, &pi.xv_shm_info);
    pi.xv_shm_info.shmid = shmget( IPC_PRIVATE, pi.xv_image->data_size, IPC_CREAT|0777);
    pi.xv_shm_info.shmaddr = shmat( pi.xv_shm_info.shmid, 0, 0);
    pi.xv_image->data = pi.xv_shm_info.shmaddr;
    pi.xv_shm_info.readOnly = 0;
    if (!XShmAttach( pi.display, &pi.xv_shm_info))
    {
      fprintf(stderr,"Cannot attach shared memory\n");
      pi.xv_image = NULL;
      return (0);
    }
    pi.xv_gc = XCreateGC( pi.display, pi.window, 0, &xgcv);
  }

  return 1;
}

void xvPut(void)
{
  XvShmPutImage( pi.display, pi.xv_port, pi.window, pi.xv_gc, pi.xv_image,
    0,0, capture->frame_width, capture->frame_height,
    0,0, capture->frame_width, capture->frame_height, False);
  XFlush( pi.display);
}

#endif

void gdkPut(void)
{
  gdk_draw_rgb_image(pi.drawable->window, pi.drawable->style->fg_gc[pi.drawable->state],
    0, 0, capture->frame_width, capture->frame_height, GDK_RGB_DITHER_MAX, 
    pi.gdk_buffer, capture->frame_width*3);
}

gint idler(gpointer p)
{
  pthread_mutex_lock(&pi.mutex);
  if (pi.buffer_is_dirty)
  {
#ifdef HAVE_X11_EXTENSIONS_XVLIB_H
    if (pi.display_method == DISPLAY_METHOD_XV)
      xvPut();
    else 
#endif
      gdkPut();
    pi.buffer_is_dirty = 0;
  }
  pthread_mutex_unlock(&pi.mutex);
  return 1;
}
